"""Built in actions for Jaseci"""

import stripe as s

from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.utils import logger
from jaseci.extens.svc.stripe_svc import StripeService
from jaseci.jsorc.live_actions import jaseci_action


def stripe() -> s:
    _stripe = JsOrc.svc("stripe", StripeService)
    if _stripe.is_running():
        return _stripe.app
    else:
        logger.info(
            "Stripe service is not running! Fallback to direct stripe call with required api key params."
        )
        return s


@jaseci_action()
def product_create(name: str, **kwargs):
    """create product"""

    return stripe().Product.create(name=name, **kwargs)


@jaseci_action()
def price_create(
    product: str,
    unit_amount: int,
    currency: str,
    **kwargs,
):
    """create product price"""

    return stripe().Price.create(
        product=product, unit_amount=unit_amount, currency=currency, **kwargs
    )


@jaseci_action()
def product_list(**kwargs):
    """retrieve all producs"""

    return stripe().Product.list(**kwargs)


@jaseci_action()
def customer_create(**kwargs):
    """create customer"""

    return stripe().Customer.create(**kwargs)


@jaseci_action()
def customer_modify(customer_id: str, **kwargs):
    """update customer"""

    return stripe().Customer.modify(customer_id, **kwargs)


@jaseci_action()
def customer_retrieve(customer_id: str, **kwargs):
    """retrieve customer information"""

    return stripe().Customer.retrieve(customer_id, **kwargs)


@jaseci_action()
def customer_delete(customer_id: str, **kwargs):
    """delete customer"""

    return stripe().Customer.delete(customer_id, **kwargs)


@jaseci_action()
def payment_method_attach(payment_method_id: str, customer_id: str, **kwargs):
    """retrieve customer list of invoices"""

    return stripe().PaymentMethod.attach(
        payment_method_id, customer=customer_id, **kwargs
    )


@jaseci_action()
def payment_method_detach(payment_method_id: str, **kwargs):
    """detach payment method from customer"""

    return stripe().PaymentMethod.detach(payment_method_id, **kwargs)


@jaseci_action()
def payment_method_list(**kwargs):
    """get customer list of payment methods"""

    return stripe().PaymentMethod.list(**kwargs)


@jaseci_action()
def invoice_create(**kwargs):
    """create customer invoice"""

    return stripe().Invoice.create(**kwargs)


@jaseci_action()
def invoice_list(**kwargs):
    """retrieve customer list of invoices"""

    return stripe().Invoice.list(**kwargs)


@jaseci_action()
def payment_intent_list(**kwargs):
    """get customer payment intents"""

    return stripe().PaymentIntent.list(**kwargs)


@jaseci_action()
def payment_intent_create(amount: int, currency: str, **kwargs):
    """Create customer payment"""

    return stripe().PaymentIntent.create(amount=amount, currency=currency, **kwargs)


@jaseci_action()
def subscription_create(customer: str, items: list, **kwargs):
    """create subcriptions"""

    return stripe().Subscription.create(customer=customer, items=items, **kwargs)


@jaseci_action()
def subscription_modify(subscription_id: str, **kwargs):
    """modify subcriptions"""

    return stripe().Subscription.modify(subscription_id, **kwargs)


@jaseci_action()
def subscription_list(**kwargs):
    """list all customer's subcriptions"""

    return stripe().Subscription.list(**kwargs)


@jaseci_action()
def subscription_retrieve(subscription_id: str, **kwargs):
    """retrieve customer subcription details"""

    return stripe().Subscription.retrieve(subscription_id, **kwargs)


@jaseci_action()
def payment_method_create(type: str, **kwargs):
    """create payment method"""

    return stripe().PaymentMethod.create(type=type, **kwargs)


@jaseci_action()
def subscription_delete(subscription_id: str, **kwargs):
    """cancel customer subscription"""

    return stripe().Subscription.delete(subscription_id, **kwargs)


@jaseci_action()
def invoice_retrieve(invoice_id: str, **kwargs):
    """get invoice information"""

    return stripe().Invoice.retrieve(invoice_id, **kwargs)


@jaseci_action()
def subscription_item_list_usage_record_summaries(subscription_item_id: str, **kwargs):
    """Create usage record"""

    return stripe().SubscriptionItem.list_usage_record_summaries(
        subscription_item_id, **kwargs
    )


@jaseci_action()
def subscription_item_create_usage_record(
    subscription_item_id: str, quantity: int, **kwargs
):
    """Create usage record"""

    return stripe().SubscriptionItem.create_usage_record(
        subscription_item_id, quantity=quantity, timestamp="now", **kwargs
    )


@jaseci_action()
def subscription_item_retrieve(subscription_item_id: str, **kwargs):
    """retrieve subscription item"""

    return stripe().SubscriptionItem.retrieve(subscription_item_id, **kwargs)


@jaseci_action()
def checkout_session_create(success_url: str, mode: str, **kwargs):
    return stripe().checkout.Session.create(
        success_url=success_url,
        mode=mode,
        **kwargs,
    )


@jaseci_action()
def billing_portal_session_create(customer: str, **kwargs):
    return stripe().billing_portal.Session.create(
        customer=customer,
        **kwargs,
    )


#################################################
#            WITH PRE CUSTOM PROCESS            #
#################################################


@jaseci_action()
def update_default_payment_method(customer_id: str, payment_method_id: str, **kwargs):
    """update default payment method of customer"""

    return customer_modify(
        customer_id,
        invoice_settings={"default_payment_method": payment_method_id},
        **kwargs,
    )


@jaseci_action()
def attach_payment_method(payment_method_id: str, customer_id: str, **kwargs):
    """attach payment method to customer"""

    paymentMethods = payment_method_list(customer=customer_id)

    paymentMethod = payment_method_attach(payment_method_id, customer_id, **kwargs)

    is_default = True
    if paymentMethods.get("data"):
        update_default_payment_method(customer_id, payment_method_id)
        is_default = False

    paymentMethod["is_default"] = is_default

    return paymentMethod


@jaseci_action()
def create_trial_subscription(
    customer_id: str,
    items: list,
    payment_method_id: str = "",
    trial_period_days: int = 14,
    **kwargs,
):
    """create customer trial subscription"""

    if payment_method_id != "":
        # attach payment method to customer
        attach_payment_method(payment_method_id, customer_id)

        # set card to default payment method
        update_default_payment_method(customer_id, payment_method_id)

    return subscription_create(
        customer_id,
        items,
        trial_period_days=trial_period_days,
        **kwargs,
    )


@jaseci_action()
def create_subscription(
    customer_id: str,
    items: list,
    payment_method_id: str = "",
    **kwargs,
):
    """create customer subscription"""

    if payment_method_id != "":
        # attach payment method to customer
        attach_payment_method(payment_method_id, customer_id)

        # set card to default payment method
        update_default_payment_method(customer_id, payment_method_id)

    return subscription_create(customer_id, items, **kwargs)


@jaseci_action()
def update_subscription_item(
    subscription_id: str, subscription_item_id: str, price_id: str, **kwargs
):
    """update subcription details"""

    return subscription_modify(
        subscription_id,
        cancel_at_period_end=False,
        items=[
            {
                "id": subscription_item_id,
                "price": price_id,
            },
        ],
        **kwargs,
    )


#################################################
#            EXECUTE CUSTOM PROCESS             #
#################################################


@jaseci_action()
def exec(api: str, *args, **kwargs):
    apis = api.split(".")

    if not apis:
        raise Exception("API is required!")

    mod = stripe()

    for api in apis:
        mod = getattr(mod, api, None)
        if not mod:
            raise Exception("Not a valid stripe API!")

    return mod(*args, **kwargs)



"""
Health APIs
"""

from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.api.interface import Interface


class HealthApi:
    """
    API for readiness probe
    """

    @Interface.public_api(allowed_methods=["get"])
    def health(self):
        """
        readiness probe
        """
        # to update
        # we might need to return some other information here
        return JsOrc.__running__



from jaseci.jsorc.live_actions import jaseci_action
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.storage_svc import StorageService


def storage():
    return JsOrc.svc("store", StorageService)


@jaseci_action()
def upload(file: str, provider: str = None, container: str = None, meta: dict = {}):
    """temp"""
    from jaseci.utils.file_handler import FileHandler

    file_handler: FileHandler = meta["h"].get_file_handler(file)

    return storage().upload(file=file_handler, provider=provider, container=container)


@jaseci_action()
def download(file: str, provider: str = None, container: str = None, meta: dict = {}):
    """temp"""
    return meta["h"].add_file_handler(storage().download(file, provider, container))


@jaseci_action()
def cdn_url(file: str, provider: str = None, container: str = None, **kwargs):
    """temp"""
    return storage().cdn_url(file, provider, container, **kwargs)


@jaseci_action()
def delete(file: str, provider: str = None, container: str = None):
    """temp"""
    return storage().delete(file, provider, container)


@jaseci_action()
def create_container(name: str, provider: str = None):
    """temp"""
    return storage().create_container(name, provider) != None


@jaseci_action()
def has_container(name: str, provider: str = None):
    """temp"""
    return storage().has_container(name, provider)


@jaseci_action()
def delete_container(name: str, provider: str = None):
    """temp"""
    return storage().delete_container(name, provider)



"""
Sentinel interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""

from jaseci.prim.architype import Architype
from jaseci.jac.interpreter.interp import Interp
from jaseci.utils.utils import parse_str_token
from jaseci.jac.ir.jac_code import jac_ast_to_ir
from jaseci.jac.machine.jac_scope import JacScope
from jaseci.prim.ability import Ability


class SentinelInterp(Interp):
    """Jac interpreter mixin for objects that will execute Jac code"""

    def run_start(self, jac_ast):
        """
        start: ver_label? element* EOF;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "ver_label":
            self.run_ver_label(kid[0])
        for i in kid:
            self.run_element(i)

    def run_ver_label(self, jac_ast):
        """
        ver_label: 'version' COLON STRING;
        """
        kid = self.set_cur_ast(jac_ast)
        self.version = parse_str_token(kid[2].token_text())

    def run_element(self, jac_ast):
        """
        element: global_var | architype | test;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "global_var":
            self.load_global_var(kid[0])
        elif kid[0].name == "architype":
            self.load_architype(kid[0])
        elif kid[0].name == "test":
            self.load_test(kid[0])

    def load_global_var(self, jac_ast):
        """
        global_var:
            KW_GLOBAL NAME EQ expression (COMMA NAME EQ expression)* SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        kid = kid[1:]
        while kid[0].name == "NAME":
            const_name = kid[0].token_text()
            if const_name in self.global_vars:
                self.rt_error(f"Global {const_name} already defined!", kid[0])
            else:
                self.run_expression(kid[2])
                self.global_vars[const_name] = self.pop().value
            kid = kid[4:] if kid[3].name == "COMMA" else kid[3:]

    def load_architype(self, jac_ast):
        """
        architype:
            KW_NODE NAME (COLON NAME)* (COLON INT)? attr_block
            | KW_EDGE NAME (COLON NAME)* attr_block
            | KW_GRAPH NAME graph_block
            | KW_ASYNC? KW_WALKER NAME namespaces? walker_block;
        """
        kid = self.set_cur_ast(jac_ast)

        is_async = kid[0].name == "KW_ASYNC" and bool(kid.pop(0))

        name = kid[1].token_text()
        kind = kid[0].token_text()
        arch = Architype(
            m_id=self._m_id,
            h=self._h,
            code_ir=jac_ast,
            name=name,
            kind=kind,
            is_async=is_async,
            parent=self,
        )

        if len(kid) > 2 and kid[2].name == "COLON":
            for i in kid[2:]:
                if i.name == "NAME":
                    arch.super_archs.append(i.token_text())
        if self.arch_ids.has_obj_by_name(arch.name, kind=arch.kind):
            self.arch_ids.destroy_obj_by_name(arch.name, kind=arch.kind)
        self.arch_ids.add_obj(arch)
        self.arch_has_preproc(kid[-1], arch)
        self.arch_can_compile(kid[-1], arch)
        return arch

    # Note: Sentinels only registers the attr_stmts

    def arch_has_preproc(self, jac_ast, arch):
        """Helper function to statically compile can stmts for arch"""
        kid = self.set_cur_ast(jac_ast)
        if jac_ast.name in ["attr_block", "walker_block"]:
            for i in kid:
                if i.name == "attr_stmt" and i.kid[0].name == "has_stmt":
                    for j in i.kid[0].kid:
                        if j.name == "has_assign":
                            has_kid = j.kid
                            is_private = False
                            is_anchor = False
                            if has_kid[0].name == "KW_PRIVATE":
                                has_kid = has_kid[1:]
                                is_private = True
                            if has_kid[0].name == "KW_ANCHOR":
                                has_kid = has_kid[1:]
                                is_anchor = True
                            var_name = has_kid[0].token_text()
                            if is_private:
                                arch.private_vars.append(var_name)
                            if is_anchor and arch.anchor_var is None:
                                arch.anchor_var = var_name
                            arch.has_vars.append(var_name)
        arch.save()

    def arch_can_compile(self, jac_ast, arch):
        """Helper function to statically compile can stmts for arch"""
        kid = self.set_cur_ast(jac_ast)
        self.push_scope(
            JacScope(parent=self, name=f"a_cgen:{jac_ast.loc_str()}", has_obj=None)
        )
        if jac_ast.name in ["attr_block", "walker_block"]:
            for i in kid:
                if i.name == "attr_stmt" and i.kid[0].name == "can_stmt":
                    self.run_can_stmt(i.kid[0], arch)
        elif jac_ast.name == "graph_block":
            self.run_can_block(jac_ast.kid[2], arch)
        self.pop_scope()

    def run_can_block(self, jac_ast, arch):
        """
        can_block: (can_stmt)*;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if i.name == "can_stmt":
                self.run_can_stmt(i, arch)

    def run_can_stmt(self, jac_ast, obj):
        """
        can_stmt:
            KW_CAN dotted_name (preset_in_out event_clause)? (
                COMMA dotted_name (preset_in_out event_clause)?
            )* SEMI
            | KW_CAN NAME event_clause? code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        kid = kid[1:]
        ir = None
        while True:
            action_type = "activity"
            access_list = None
            preset_in_out = None
            if kid[0].name == "NAME":
                action_name = kid[0].token_text()
            else:
                action_name = self.run_dotted_name(kid[0])
            kid = kid[1:]
            if len(kid) > 0 and kid[0].name == "preset_in_out":
                preset_in_out = jac_ast_to_ir(kid[0])
                kid = kid[1:]
            if len(kid) > 0 and kid[0].name == "event_clause":
                action_type, access_list = self.run_event_clause(kid[0])
                kid = kid[1:]
            if kid[0].name != "code_block":
                self.check_builtin_action(action_name, jac_ast)
            else:
                ir = kid[0]
            getattr(obj, f"{action_type}_ability_ids").add_obj(
                Ability(
                    m_id=self._m_id,
                    h=self._h,
                    name=action_name,
                    kind="ability",
                    code_ir=ir,
                    preset_in_out=preset_in_out,
                    access_list=access_list,
                    parent=self,
                )
            )
            if not len(kid) or kid[0].name != "COMMA":
                break
            else:
                kid = kid[1:]

    def run_event_clause(self, jac_ast):
        """
        event_clause:
                KW_WITH name_list? (KW_ENTRY | KW_EXIT | KW_ACTIVITY);
        """
        kid = self.set_cur_ast(jac_ast)
        nl = []
        if kid[1].name == "name_list":
            nl = self.run_name_list(kid[1])
        return kid[-1].token_text(), nl

    def load_test(self, jac_ast):
        """
        test:
            KW_TEST NAME? multistring KW_WITH (
                graph_ref
                | KW_GRAPH graph_block
            ) KW_BY (
                (walker_ref spawn_ctx? (code_block | SEMI))
                | KW_WALKER walker_block
            );
        """
        kid = self.set_cur_ast(jac_ast)
        (
            self.run_multistring(kid[2])
            if kid[1].name == "NAME"
            else self.run_multistring(kid[1])
        )
        testcase = {
            "name": kid[1].token_text() if kid[1].name == "NAME" else "",
            "title": self.pop().value,
            "graph_ref": None,
            "graph_block": None,
            "walker_ref": None,
            "spawn_ctx": None,
            "assert_block": None,
            "walker_block": None,
            "outcome": None,
        }
        kid = kid[4:] if kid[1].name == "NAME" else kid[3:]
        if kid[0].name == "graph_ref":
            graph_name = kid[0].kid[-1].token_text()
            if not self.arch_ids.has_obj_by_name(graph_name, kind="graph"):
                self.rt_error(f"Graph {graph_name} not found!", kid[0])
            testcase["graph_ref"] = graph_name
        else:
            kid = kid[1:]
            testcase["graph_block"] = jac_ast_to_ir(kid[0])
        kid = kid[2:]
        if kid[0].name == "walker_ref":
            walker_name = kid[0].kid[-1].token_text()
            if not self.arch_ids.has_obj_by_name(name=walker_name, kind="walker"):
                self.rt_error(f"Walker {walker_name} not found!", kid[0])
            testcase["walker_ref"] = walker_name
            kid = kid[1:]
            if kid[0].name == "spawn_ctx":
                testcase["spawn_ctx"] = jac_ast_to_ir(kid[0])
                kid = kid[1:]
            if kid[0].name == "code_block":
                testcase["assert_block"] = jac_ast_to_ir(kid[0])
        else:
            kid = kid[1:]
            testcase["walker_block"] = jac_ast_to_ir(kid[0])

        self.testcases.append(testcase)



"""
Sentinel interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""

from jaseci.prim.node import Node
from jaseci.prim.edge import Edge
from jaseci.prim.walker import Walker
from jaseci.jac.interpreter.interp import Interp
from jaseci.jac.machine.jac_scope import JacScope
from jaseci.jac.machine.jac_value import JacValue


class ArchitypeInterp(Interp):
    """Jac interpreter mixin for objects that will execute Jac code"""

    def run_architype(self, jac_ast):
        """
        architype:
            KW_NODE NAME (COLON NAME)* attr_block
            | KW_EDGE NAME (COLON NAME)* attr_block
            | KW_TYPE NAME struct_block
            | KW_GRAPH NAME graph_block
            | KW_ASYNC? KW_WALKER NAME namespaces? walker_block;
        """
        if jac_ast is None:  # Using defaults
            if self.kind == "node" and self.name in ["root", "generic"]:
                return Node(
                    m_id=self._m_id,
                    h=self._h,
                    kind=self.kind,
                    name=self.name,
                    parent=self.parent(),
                )
            elif self.kind == "edge" and self.name in ["generic"]:
                return Edge(
                    m_id=self._m_id,
                    h=self._h,
                    kind=self.kind,
                    name=self.name,
                    parent=self.parent(),
                )

        kid = self.set_cur_ast(jac_ast)

        self.push_scope(JacScope(parent=self, name=f"spawn:{jac_ast.loc_str()}"))
        if kid[0].name == "KW_NODE":
            item = Node(
                m_id=self._m_id,
                h=self._h,
                kind=kid[0].token_text(),
                name=kid[1].token_text(),
                parent=self.parent(),
            )
            self.build_object_with_supers(item, kid[-1])
        elif kid[0].name == "KW_EDGE":
            item = Edge(
                m_id=self._m_id,
                h=self._h,
                kind=kid[0].token_text(),
                name=kid[1].token_text(),
                parent=self.parent(),
            )
            self.build_object_with_supers(item, kid[-1])
        elif kid[0].name == "KW_TYPE":
            item = self.run_struct_block(kid[-1])
        elif kid[0].name == "KW_GRAPH":
            item = self.run_graph_block(kid[-1])
        elif kid[0].name == "KW_WALKER":
            item = Walker(
                m_id=self._m_id,
                h=self._h,
                code_ir=jac_ast,
                name=kid[1].token_text(),
                kind=kid[0].token_text(),
                parent=self.parent(),
                is_async=self.is_async,
            )
            if kid[2].name == "namespaces":
                item.namespaces = self.run_namespaces(jac_ast.kid[2])
            self.build_object_with_supers(item, kid[-1])
        elif jac_ast.name == "graph_block":  # usedi n jac tests
            item = self.run_graph_block(jac_ast)
        self.pop_scope()
        return item

    def run_namespaces(self, jac_ast):
        """
        namespaces: COLON name_list;
        """
        return self.run_name_list(jac_ast.kid[1])

    def run_walker_block(self, jac_ast, obj):
        """
        walker_block:
            LBRACE attr_stmt* walk_entry_block? (
                statement
                | walk_activity_block
            )* walk_exit_block? RBRACE;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if i.name == "attr_stmt":
                self.run_attr_stmt(jac_ast=i, obj=obj)

    def run_attr_block(self, jac_ast, obj):
        """
        attr_block:
            LBRACE (attr_stmt)* RBRACE
            | COLON (attr_stmt)* SEMI
            | SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if i.name == "attr_stmt":
                self.run_attr_stmt(i, obj)

    def run_attr_stmt(self, jac_ast, obj):
        """
        attr_stmt: has_stmt | can_stmt;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "has_stmt":
            self.run_has_stmt(kid[0], obj)
        #  Can statements in architype handled in architype load

    def run_has_stmt(self, jac_ast, obj):
        """
        has_stmt: KW_HAS has_assign (COMMA has_assign)* SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if i.name == "has_assign":
                self.run_has_assign(i, obj)

    def run_has_assign(self, jac_ast, obj):
        """
        has_assign: KW_PRIVATE? KW_ANCHOR? (NAME | NAME EQ expression);
        """
        kid = self.set_cur_ast(jac_ast)
        while kid[0].name in ["KW_PRIVATE", "KW_ANCHOR"]:
            kid = kid[1:]
        var_name = kid[0].token_text()
        var_val = None  # jac's null
        if len(kid) > 1:
            self.run_expression(kid[2])
            var_val = self.pop().value
        if isinstance(obj, dict):
            obj[var_name] = var_val
        # Runs only once for walkers
        elif var_name not in obj.context.keys() or obj.j_type != "walker":
            JacValue(
                self, ctx=obj, name=var_name, value=var_val, create_mode=True
            ).write(kid[0], force=True)

    def run_struct_block(self, jac_ast):
        """
        struct_block: LBRACE (has_stmt)* RBRACE | COLON has_stmt | SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = {}
        for i in kid:
            if i.name == "has_stmt":
                self.run_has_stmt(i, ret)
        return ret

    def run_graph_block(self, jac_ast):
        """
        graph_block:
            LBRACE has_root can_block KW_SPAWN code_block RBRACE
            | COLON has_root can_block KW_SPAWN code_block SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        root_name = self.run_has_root(kid[1])
        try:
            for ability in self.get_all_abilities().obj_list():
                self._jac_scope.add_action(ability)

            self.run_code_block(kid[4])
        except Exception as e:
            self.rt_error(f"Internal Exception: {e}", self._cur_jac_ast)
        local_state = self._jac_scope.local_scope
        if root_name in local_state.keys():
            obj = local_state[root_name]
            if not isinstance(obj, Node):
                self.rt_error(f"{root_name} is {type(obj)} not node!", kid[3])
            return obj
        else:
            self.rt_error("Graph didn't produce root node!", kid[3])

    def run_has_root(self, jac_ast):
        """
        has_root: KW_HAS KW_ANCHOR NAME SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        return kid[2].token_text()

    # Helper Functions ##################

    def build_object_with_supers(self, item, jac_ast):
        for i in self.super_archs:
            super_jac_ast = (
                self.parent()
                .arch_ids.get_obj_by_name(name=i, kind=item.kind)
                .get_jac_ast()
                .kid[-1]
            )
            (
                self.run_attr_block(super_jac_ast, item)
                if not isinstance(item, Walker)
                else self.run_walker_block(super_jac_ast, item)
            )
        (
            self.run_attr_block(jac_ast, item)
            if not isinstance(item, Walker)
            else self.run_walker_block(jac_ast, item)
        )



"""
General action base class with automation for hot loading
"""

from jaseci.utils.utils import logger, ColCodes as Cc
from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from pydantic import validate_arguments
from time import time
import inspect
import uvicorn
import os
import re

var_args = re.compile(r"^\*[^\*]")
var_kwargs = re.compile(r"^\*\*[^\*]")

args_kwargs = (2, 4)
remote_actions = {}
registered_apis = []
registered_endpoints = []
ACTIONS_SPEC_LOC = "/jaseci_actions_spec/"
JS_ACTION_PREAMBLE = "js_action_"
JS_ENDPOINT_PREAMBLE = "js_endpoint_"


def mark_as_remote(api):
    registered_apis.append(api)


def mark_as_endpoint(endpoint):
    registered_endpoints.append(endpoint)


def serv_actions():
    """Returns fastAPI app interface for actions"""
    app = FastAPI(
        title="Jaseci Action Set API",
        description="A Jaseci Action Set",
    )

    @app.get("/")
    def home_redirect():
        return RedirectResponse(url="/docs")

    @app.get(ACTIONS_SPEC_LOC)
    def action_list():
        return remote_actions

    for i in registered_apis:
        gen_api_service(app, *i)
    for i in registered_endpoints:
        gen_endpoint(app, *i)
    return app


def gen_api_service(app, func, act_group, aliases, caller_globals):
    """Helper for jaseci_action decorator"""
    # Construct list of action apis available
    act_group = (
        [os.path.basename(inspect.getfile(func)).split(".")[0]]
        if act_group is None
        else act_group
    )
    varnames = []
    for param in inspect.signature(func).parameters.values():
        varnames.append(str(param) if param.kind in args_kwargs else param.name)

    remote_actions[f"{'.'.join(act_group+[func.__name__])}"] = varnames

    # Need to get pydatic model for func signature for fastAPI post
    model = validate_arguments(func).model

    # Keep only fields present in param list in base model
    keep_fields = {}
    for name in varnames:
        if var_args.match(name):
            keep_fields[name] = model.__fields__[name[1:]]
            keep_fields[name].name = name
            keep_fields[name].alias = name
        elif var_kwargs.match(name):
            keep_fields[name] = model.__fields__[name[2:]]
            keep_fields[name].name = name
            keep_fields[name].alias = name
        else:
            field = model.__fields__.get(name)
            if field:
                keep_fields[name] = field
    model.__fields__ = keep_fields

    # Create duplicate funtion for api endpoint and inject in call site globals
    @app.post(f"/{func.__name__}/")
    def new_func(params: model = model.construct()):
        params: dict = params.__dict__
        pl_peek = str(params)[:128]
        logger.info(str(f"Incoming call to {func.__name__} with {pl_peek}"))
        start_time = time()

        args = []
        kwargs = {}
        fp1 = inspect.signature(func).parameters.values()
        fp2 = list(fp1)

        # try to process args
        for param in fp1:
            _param = str(param) if param.kind in args_kwargs else param.name
            if _param in params:
                fp2.remove(param)
                if var_args.match(_param):
                    for arg in params.get(_param) or []:
                        args.append(arg)
                    break
                elif var_kwargs.match(_param):
                    kwargs.update(params.get(_param) or {})
                    break
                args.append(params.get(_param))
            else:
                break

        # try to process kwargs
        for param in fp2:
            param = str(param) if param.kind in args_kwargs else param.name
            if param in params:
                if var_kwargs.match(param):
                    kwargs.update(params.get(param) or {})
                    break
                kwargs[param] = params.get(param)
            else:
                break

        ret = validate_arguments(func)(*args, **kwargs)
        tot_time = time() - start_time
        logger.info(
            str(
                f"API call to {Cc.TG}{func.__name__}{Cc.EC}"
                f" completed in {Cc.TY}{tot_time:.3f} seconds{Cc.EC}"
            )
        )
        return ret

    if func.__name__ == "setup":
        new_func = app.on_event("startup")(new_func)
    for i in aliases:
        new_func = app.post(f"/{i}/")(new_func)
        remote_actions[f"{'.'.join(act_group+[i])}"] = varnames
    caller_globals[f"{JS_ACTION_PREAMBLE}{func.__name__}"] = new_func


def gen_endpoint(app, func, endpoint, mount, caller_globals):
    """Helper for jaseci_action decorator"""
    # Create duplicate funtion for api endpoint and inject in call site globals
    if mount is not None:
        app.mount(*mount)
    caller_globals[f"{JS_ENDPOINT_PREAMBLE}{func.__name__}"] = app.get(endpoint)(func)


def launch_server(port=80, host="0.0.0.0"):
    uvicorn.run(serv_actions(), port=port, host=host)



"""
Object sharing APIs
"""
from jaseci.extens.api.interface import Interface
from jaseci.prim.element import Element


class ShareApi:
    """
    Sharing APIs.
    """

    def __init__(self):
        self.incoming = {}

    @Interface.private_api()
    def share_object(
        self, receiver: str, objs: list[Element] = [], read_only: bool = True
    ):
        """
        Sharing an object with a user
        obj: The list of elements to share
        receiver: master of the receiving user
        read_only: if set true, the object shared will be shared as read-only
        """
        # Get the master object by id
        receiver_mast = self._h.get_obj(
            caller_id=self.jid, item_id=receiver, override=True
        )

        for obj in objs:
            # Grant read-only permission to the new user
            self.object_perms_grant(obj, receiver_mast, read_only=read_only)
            obj.save()

            # Add the objet id to the receiver's incoming list
            receiver_mast.incoming[str(obj.id)] = {"jid": self.jid, "name": self.name}

        receiver_mast.save()
        self.save()

        return {
            "objects": [str(obj) for obj in objs],
            "sharer": str(self),
            "receiver": str(receiver_mast),
        }

    @Interface.private_api()
    def share_incoming_pop(self, obj_id: str):
        """
        Remove an item from the incoming list
        """
        try:
            self.incoming.pop(obj_id)
            return True
        except Exception:
            return None

    @Interface.private_api()
    def share_incoming_list(self):
        """
        Get the incoming objects
        """
        return self.incoming



from jaseci.utils.test_core import CoreTest


class UncommonImplementationTest(CoreTest):
    """Unit tests for Uncommon Implementaion APIs"""

    fixture_src = __file__

    def test_overrided_action_ability(self):
        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("action_override.jac")}],
        )

        reports = self.call(self.smast, ["walker_run", {"name": "check"}])["report"]
        keys = [
            "log",
            "out",
            "js_input",
            "input",
            "js_round",
            "round",
            "err",
            "sleep",
            "sort_by_col",
            "time_now",
            "set_global",
            "get_global",
            "actload_local",
            "actload_remote",
            "actload_module",
            "destroy_global",
            "set_perms",
            "get_perms",
            "grant_perms",
            "revoke_perms",
            "get_report",
            "clear_report",
            "log_activity",
        ]
        for key in keys:
            # node tester - testing first report: std should be the default
            self.assertIsInstance(reports[0][key], dict)
            self.assertEqual(reports[0][key]["name"], f"std.{key}")
            self.assertEqual(reports[0][key]["kind"], "ability")

            # walker check - tester report: std should be the default
            self.assertIsInstance(reports[3][key], dict)
            self.assertEqual(reports[3][key]["name"], f"std.{key}")
            self.assertEqual(reports[3][key]["kind"], "ability")

            # node tester - testing after std.set_global = 1
            # all actions should be the same except std.set_global
            if key != "set_global":
                self.assertIsInstance(reports[1][key], dict)
                self.assertEqual(reports[1][key]["name"], f"std.{key}")
                self.assertEqual(reports[1][key]["kind"], "ability")

        self.assertIsInstance(reports[1]["set_global"], int)
        self.assertEqual(reports[1]["set_global"], 1)

        # node tester - testing third report: std set to 1
        self.assertEqual(reports[2], 1)

    def test_node_ability_from_global_sentinel(self):
        self.call(self.mast, ["graph_create", {}])

        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("uncommon_impl.jac")}],
        )

        self.call(self.smast, ["global_sentinel_set", {}])
        self.call(self.mast, ["sentinel_active_global", {}])

        res = self.call(self.smast, ["walker_run", {"name": "check_ability"}])
        self.assertEqual("super_master", res["report"][0]["j_type"])
        self.assertEqual([], res["report"][1])

        # since admin already create `with_ability` node on init,
        # it should not have `inner_node`
        res = self.call(self.smast, ["walker_run", {"name": "check_ability"}])
        self.assertEqual("super_master", res["report"][0]["j_type"])
        self.assertEqual(0, len(res["report"][1]))

        # users should be able to access their inner nodes inside ability
        res = self.call(self.mast, ["walker_run", {"name": "check_ability"}])
        # before it will return the owner of the sentinel which is the super_master
        self.assertEqual("master", res["report"][0]["j_type"])
        self.assertEqual(1, len(res["report"][1]))

    def test_existing_node_in_graph_but_not_in_sentinel(self):
        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("with_node_b.jac")}],
        )

        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("without_node_b.jac")}],
        )

        res = self.call(self.smast, ["walker_run", {"name": "init"}])

        self.assertTrue(res["success"])
        self.assertEqual(1, len(res["report"]))
        self.assertTrue("a", len(res["report"][0]["name"]))



"""Built in actions for Jaseci"""

import requests
from jaseci.jsorc.live_actions import jaseci_action


@jaseci_action()
def get(url: str, data: dict = {}, header: dict = {}):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.get(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def post(url: str, data: dict = {}, header: dict = {}):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.post(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def put(url: str, data: dict = {}, header: dict = {}):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.put(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def delete(url: str, data: dict = {}, header: dict = {}):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.delete(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def head(url: str, data: dict = {}, header: dict = {}):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.head(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def options(url: str, data: dict = {}, header: dict = {}):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.options(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def multipart(
    url: str, data: dict = {}, files: list = [], header: dict = {}, meta: dict = {}
):
    """
    Issue request
    Param 1 - url
    Param 2 - data (Optional) used for request body
    Param 3 - files (Optional) used for request files
    Param 4 - header

    Return - response object
    """

    hook = meta["h"]

    if not files:
        return {
            "status_code": 400,
            "error": "Please include base64 using this format "
            '{"field":val,"name":val,"base64":val} '
            "using parameter `file` and `files` for array file",
        }

    _files = []
    stream_to_be_close = []

    if files is not None:
        for f in files:
            file_handler = hook.get_file_handler(f)
            stream = file_handler.open("rb", None, True)
            stream_to_be_close.append(stream)
            _files.append(
                (
                    file_handler.field or "file",
                    (file_handler.name, stream, file_handler.content_type),
                )
            )

    res = requests.post(url, data=data, files=_files, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text

    for stream in stream_to_be_close:
        stream.close()

    return ret



import re
import mimetypes
from os import unlink
from io import BytesIO
from uuid import uuid4
from base64 import b64encode
from tempfile import _TemporaryFileWrapper
from os.path import basename, exists
from json import load, dump


filename_from_disposition = re.compile(r"filename=(['\"])(.*?)\1")


class FileHandler:
    ###################################################
    #                     BUILDER                     #
    ###################################################

    def __init__(
        self,
        name: str = None,
        content_type: str = None,
        field: str = None,
        persist: bool = False,
    ):
        self.id = str(uuid4())
        self.name = name
        self.content_type = content_type or mimetypes.guess_type(name)[0]
        self.field = field
        self.absolute_name = f"{self.id}-{name}"
        self.absolute_path = f"/tmp/{self.absolute_name}"
        self.persist = persist
        self.buffer = None

    @classmethod
    def fromTemporaryFileWrapper(
        cls,
        file: _TemporaryFileWrapper,
        name: str,
        content_type: str,
        field: str = None,
    ):
        file_handler = cls(name, content_type, field)
        file_handler.absolute_path = file.name

        file._closer.delete = False
        file.close()

        return file_handler

    @classmethod
    def fromBytesIO(
        cls, file: BytesIO, name: str, content_type: str, field: str = None
    ):
        file_handler = cls(name, content_type, field)

        with open(file_handler.absolute_path, "wb") as f:
            f.write(file.getbuffer())
        file.close()

        return file_handler

    @classmethod
    def fromRequest(cls, content: bytes, content_disposition: str):
        name = "tmp"
        searcher = filename_from_disposition.search(content_disposition)
        if len(searcher.groups()):
            name = searcher.group(2)

        file_handler = cls(name)

        with open(file_handler.absolute_path, "wb") as f:
            f.write(content)

        return file_handler

    @classmethod
    def fromPath(cls, path: str):
        name = basename(path)

        file_handler = cls(name, persist=True)
        file_handler.absolute_name = name
        file_handler.absolute_path = path

        return file_handler

    ###################################################
    #                     CONTROL                     #
    ###################################################

    def read(self, offset: int = None, mode: str = "r", encoding: str = None) -> str:
        if self.buffer:
            if offset != None:
                self.buffer.seek(offset)
            content = self.buffer.read()
        else:
            self.open(mode, encoding)
            content = self.buffer.read()
            self.close()

        return content

    def open(
        self, mode: str = "r", encoding: str = None, detached: bool = False, **kwargs
    ):
        if detached:
            return open(self.absolute_path, mode, encoding=encoding, **kwargs)

        if not self.buffer:
            self.buffer = open(self.absolute_path, mode, encoding=encoding, **kwargs)

    def write(self, content: str):
        if self.buffer:
            self.buffer.write(content)

    def seek(self, offset: int, whence: int = 0):
        if self.buffer:
            self.buffer.seek(offset, whence)

    def flush(self):
        if self.buffer:
            self.buffer.flush()

    def close(self):
        if self.buffer:
            self.buffer.close()
            self.buffer = None

    def delete(self):
        self.close()
        try:
            unlink(self.absolute_path)
        except Exception:
            pass

    def is_open(self) -> bool:
        return self.buffer != None

    def exists(self) -> bool:
        return exists(self.absolute_path)

    def attr(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
            "content_type": self.content_type,
            "field": self.field,
            "absolute_name": self.absolute_name,
            "absolute_path": self.absolute_path,
            "persist": self.persist,
        }

    ###################################################
    #                     UTILITY                     #
    ###################################################

    def to_bytes(self):
        with self.open("rb", detached=True) as buffer:
            return buffer.read()

    def to_base64(self):
        return b64encode(self.to_bytes()).decode()

    def to_json(self):
        with self.open(detached=True) as buffer:
            return load(buffer)

    def dump_json(self, json, indent: int = None):
        with self.open("w", detached=True) as buffer:
            dump(json, buffer, indent=indent)



from jaseci.utils.test_core import CoreTest, jac_testcase


class MathsTests(CoreTest):
    """UnitTest for Regex Module"""

    fixture_src = __file__

    @jac_testcase("maths.jac", "ceil_test")
    def test_ceil(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 124)

    @jac_testcase("maths.jac", "comb_test")
    def test_comb(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 4950)

    @jac_testcase("maths.jac", "copysign_test")
    def test_copysign(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, -100.0)

    @jac_testcase("maths.jac", "fabs_test")
    def test_fabs(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 10.0)

    @jac_testcase("maths.jac", "factorial_test")
    def test_factorial(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 3628800)

    @jac_testcase("maths.jac", "floor_test")
    def test_floor(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 10)

    @jac_testcase("maths.jac", "fmod_test")
    def test_fmod(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 1.0)

    @jac_testcase("maths.jac", "frexp_test")
    def test_frexp(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, [0.78125, 7])

    @jac_testcase("maths.jac", "fsum_test")
    def test_fsum(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 18.0)

    @jac_testcase("maths.jac", "gcd_test")
    def test_gcd(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 1)

    @jac_testcase("maths.jac", "isclose_test")
    def test_isclose(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, False)

    @jac_testcase("maths.jac", "isfinite_test")
    def test_isfinite(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, True)

    @jac_testcase("maths.jac", "isinf_test")
    def test_isinf(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, False)

    @jac_testcase("maths.jac", "isnan_test")
    def test_isnan(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, False)

    @jac_testcase("maths.jac", "isqrt_test")
    def test_isqrt(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 3)

    @jac_testcase("maths.jac", "lcm_test")
    def test_lcm(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 420)

    @jac_testcase("maths.jac", "ldexp_test")
    def test_isqrt(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 36.0)

    @jac_testcase("maths.jac", "modf_test")
    def test_modf(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, [0.0, 9.0])

    @jac_testcase("maths.jac", "nextafter_test")
    def test_nextafter(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 8.999999999999998)

    @jac_testcase("maths.jac", "perm_test")
    def test_perm(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 25852016738884976640000)

    @jac_testcase("maths.jac", "prod_test")
    def test_prod(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 120)

    @jac_testcase("maths.jac", "remainder_test")
    def test_remainder(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 3.0)

    @jac_testcase("maths.jac", "trunc_test")
    def test_trunc(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 23)

    @jac_testcase("maths.jac", "trunc_test")
    def test_trunc(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 23)

    @jac_testcase("maths.jac", "ulp_test")
    def test_ulp(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 3.552713678800501e-15)

    @jac_testcase("maths.jac", "cubert_test")
    def test_cubert(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 3.0)

    @jac_testcase("maths.jac", "exp_test")
    def test_exp(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 532048240601.79865)

    @jac_testcase("maths.jac", "log_test")
    def test_log(self, ret):
        ret = ret["report"][0]
        self.assertEqual(ret, 1.4313637641589871)



"""Built in actions for Jaseci"""

from datetime import datetime, timedelta
from jaseci.jsorc.live_actions import jaseci_action
from dateutil.relativedelta import relativedelta


@jaseci_action()
def quantize_to_year(date: str):
    date = datetime.fromisoformat(date)
    date = date.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
    return date.isoformat()


@jaseci_action()
def quantize_to_month(date: str):
    date = datetime.fromisoformat(date)
    date = date.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    return date.isoformat()


@jaseci_action()
def quantize_to_week(date: str):
    date = datetime.fromisoformat(date)
    date = date.replace(hour=0, minute=0, second=0, microsecond=0)
    date = date - timedelta(days=date.weekday())
    return date.isoformat()


@jaseci_action()
def quantize_to_day(date: str):
    date = datetime.fromisoformat(date)
    date = date.replace(hour=0, minute=0, second=0, microsecond=0)
    return date.isoformat()


@jaseci_action()
def datetime_now():
    """Get utc date time now in iso format"""
    return datetime.utcnow().isoformat()


@jaseci_action()
def date_now():
    """Get utc date now in iso format"""
    return datetime.utcnow().date().isoformat()


@jaseci_action()
def timestamp_now():
    """Get utc date time for now in iso format"""
    return int(datetime.utcnow().timestamp())


@jaseci_action()
def datetime_obj(date: str = None):
    """Get utc date time now or parse one to dict format"""
    if date:
        date = datetime.fromisoformat(date)
    else:
        date = datetime.utcnow()

    return {
        "year": date.year,
        "month": date.month,
        "weekday": date.isoweekday(),
        "day": date.day,
        "hour": date.hour,
        "minute": date.minute,
        "second": date.second,
        "microsecond": date.microsecond,
        "iso": date.isoformat(),
    }


@jaseci_action()
def datetime_add(date: str = None, **kwargs):
    """
    Get utc date time now or parse one then add timedelta
    kwargs: this should be valid timedelta arguments
    """
    if date:
        date = datetime.fromisoformat(date)
    else:
        date = datetime.utcnow()

    return (date + relativedelta(**kwargs)).isoformat()


@jaseci_action()
def datetime_sub(date: str = None, **kwargs):
    """
    Get utc date time now or parse one then subtract timedelta
    kwargs: this should be valid timedelta arguments
    """
    if date:
        date = datetime.fromisoformat(date)
    else:
        date = datetime.utcnow()

    return (date - relativedelta(**kwargs)).isoformat()


@jaseci_action()
def date_day_diff(start_date: str, end_date: str = None):
    # Try to deal with some incompatible date in old nodes
    # Due date has not been saving in isoformat so this doesn't work
    # for example,'2021-03-31T04:00:00.000Z'
    start_date = datetime.fromisoformat(start_date.split("T")[0])

    if end_date is None:
        end_date = datetime.utcnow()
    else:
        end_date = datetime.fromisoformat(end_date.split("T")[0])

    delta = end_date - start_date

    return delta.days



from datetime import datetime, timedelta

from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.file_handler import FileHandler

from libcloud.storage.types import Provider, ObjectDoesNotExistError
from libcloud.storage.providers import get_driver as _get_driver
from libcloud.storage.base import StorageDriver, Object, Container
from libcloud.storage.drivers.azure_blobs import AzureBlobsStorageDriver

from azure.storage.blob import generate_blob_sas


@JsOrc.service(
    name="store",
    config="STORE_CONFIG",
    manifest="STORE_MANIFEST",
    priority=0,
    proxy=True,
)
class StorageService(JsOrc.CommonService):
    def run(self):
        self.app = StorageHandler(self.config["providers"], self.config.get("default"))

        # Testing
        # obj = self.upload(FileHandler.fromPath("test.txt"))
        # self.download(obj["name"])

    def upload(
        self, file: FileHandler, provider: str = None, container: str = None
    ) -> dict:
        provider = self.app.get_storage(provider)
        buffer = file.open(mode="rb", encoding=None, detached=True)

        metadata = {
            "id": file.id,
            "name": file.name,
            "created": datetime.utcnow().isoformat(),
        }

        # metadata doesn't support empty value
        if file.content_type:
            metadata["content_type"] = file.content_type

        # metadata doesn't support empty value
        if file.field:
            metadata["field"] = file.field

        if container:
            container = provider["driver"].get_container(container_name=container)

        obj: Object = provider["driver"].upload_object_via_stream(
            iterator=buffer,
            container=container or provider["container"],
            object_name=file.absolute_name,
            extra={"meta_data": metadata},
        )

        buffer.close()

        return obj.get_cdn_url()

    def download(
        self, file: str, provider: str = None, container: str = None
    ) -> FileHandler:
        provider = self.app.get_storage(provider)

        obj: Object = provider["driver"].get_object(
            container_name=container or provider["container"].name, object_name=file
        )

        temp_file = FileHandler(file)
        temp_file.open(mode="wb", encoding=None)
        for chunk in provider["driver"].download_object_as_stream(obj):
            temp_file.write(chunk)
        temp_file.close()

        return temp_file

    def cdn_url(
        self, file: str, provider: str = None, container: str = None, **kwargs
    ) -> str:
        provider = self.app.get_storage(provider)
        driver = provider["driver"]

        return driver.get_object_cdn_url(
            driver.get_object(
                container_name=container or provider["container"].name, object_name=file
            ),
            **kwargs,
        )

    def delete(self, file: str, provider: str = None, container: str = None) -> bool:
        provider = self.app.get_storage(provider)

        try:
            obj: Object = provider["driver"].get_object(
                container_name=container or provider["container"].name, object_name=file
            )
        except ObjectDoesNotExistError:
            return True

        return provider["driver"].delete_object(obj)

    def create_container(self, name: str, provider: str = None) -> Container:
        provider = self.app.get_storage(provider)

        return provider["driver"].create_container(container_name=name)

    def has_container(self, name: str, provider: str = None) -> bool:
        provider = self.app.get_storage(provider)

        return isinstance(
            provider["driver"].get_container(container_name=name), Container
        )

    def delete_container(self, name: str, provider: str = None) -> bool:
        provider = self.app.get_storage(provider)

        return provider["driver"].delete_container(
            provider["driver"].get_container(container_name=name)
        )


class StorageHandler:
    def __init__(self, providers: dict, default: str = None):
        self.default = default
        self.storage = {}

        for key, value in providers.items():
            try:
                driver = get_driver(value["provider"])(**value["credentials"])
                container = driver.get_container(container_name=value["container"])
                self.storage[key] = {
                    "source": value,
                    "driver": driver,
                    "container": container,
                }
                self.storage[key]
            except Exception as e:
                self.storage[key] = {"source": value, "error": str(e)}

    def get_storage(self, provider: str = None) -> StorageDriver:
        if provider:
            return self.storage[provider]
        elif self.default:
            return self.storage[self.default]


#############################################################################################
# ----------------------------------- PROVIDER OVERRIDE ----------------------------------- #
#############################################################################################


class AzureBlobsStorageDriverOverride(AzureBlobsStorageDriver):
    AZURE_STORAGE_CDN_URL_DATE_FORMAT = "%Y-%m-%dT%HH:%M:%SZ"

    def __init__(
        self,
        key,
        secret=None,
        **kwargs,
    ):
        self.secret_raw = secret
        super().__init__(key, secret, **kwargs)

    def get_object_cdn_url(self, obj: Object, ex_expiry: float = 24, **kwargs):
        sas = generate_blob_sas(
            self.key,
            obj.container.name,
            obj.name,
            account_key=self.secret_raw,
            permission="r",
            expiry=datetime.utcnow() + timedelta(hours=ex_expiry),
            **kwargs,
        )

        return "{scheme}://{host}:{port}{action}?{sas_token}".format(
            scheme="https" if self.secure else "http",
            host=self.connection.host,
            port=self.connection.port,
            action=self.connection.morph_action_hook(
                self._get_object_path(obj.container, obj.name)
            ),
            sas_token=sas,
        )


overrided_providers = {Provider.AZURE_BLOBS: AzureBlobsStorageDriverOverride}


def get_driver(provider):
    return overrided_providers.get(provider) or _get_driver(provider)



"""
Webhook API
"""

from jaseci.extens.api.interface import Interface
from fastapi import HTTPException
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.stripe_svc import StripeService

import stripe as _stripe


class WebhookApi:
    """
    Webhook API
    """

    @Interface.public_api(url_args=["provider"], allowed_methods=["post"])
    def webhook(self, provider: str, _req_ctx: dict = {}):
        """Handle webhook logic"""
        req_body = _req_ctx["body"]

        if provider == "stripe":
            stripe_service = JsOrc.svc("stripe", StripeService)
            stripe = stripe_service.poke(_stripe)

            # to be updated
            stripe_service.get_event(
                self._h.get_file_handler(_req_ctx["raw"]).to_bytes(),
                _req_ctx["headers"],
            )

            payload_obj = req_body.get("data").get("object")
            payload_meta = payload_obj.get("metadata")

            customer_id = payload_obj.get("customer")
            customer_meta = (
                stripe.Customer.retrieve(id=customer_id).get("metadata")
                if customer_id
                else {}
            )

            master_id = payload_meta.get("master") or customer_meta.get("master")
            master = self._h.get_obj(master_id, master_id)

            node_id = payload_meta.get("node") or customer_meta.get("node")
            if not node_id:
                node_id = master.active_gph_id
            node = self._h.get_obj(master_id, node_id)

            sentinel_id = (
                payload_meta.get("sentinel")
                or customer_meta.get("sentinel")
                or master.active_snt_id
                or self._h.get_glob("GLOB_SENTINEL")
            )
            sentinel = self._h.get_obj(
                master_id,
                (
                    self._h.get_glob("GLOB_SENTINEL")
                    if sentinel_id == "global"
                    else sentinel_id
                ),
            )

            payload = {"event": req_body}
            self.seek_committer(master)

            wlk = stripe_service.get_walker(req_body["type"])

            return master.walker_run(
                name=wlk, nd=node, ctx=payload, _req_ctx=_req_ctx, snt=sentinel
            )
        else:
            raise HTTPException(
                status_code=400, detail=str(provider + " webhook is not yet supported")
            )



from jaseci.utils.test_core import CoreTest
from unittest.mock import patch
from io import BufferedReader


class FileTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def setUp(self):
        super().setUp()

    def tearDown(self):
        self.mast._h.clean_file_handler()
        super().tearDown()

    def call_walker(self, name, report=True):
        res = self.call(
            self.mast,
            [
                "walker_run",
                {"name": name, "ctx": {}},
            ],
        )

        self.assertTrue(res["success"])
        return res["report"] if report else res

    def test_file_syntax(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("file.jac")}],
        )

        reps = self.call_walker("new_file")
        self.assertEqual("", reps[1])
        self.assertEqual("", reps[2])
        self.assertEqual("0", reps[3])
        self.assertEqual("00", reps[4])
        self.assertEqual("MDA=", reps[5])
        self.assertTrue(reps[6])
        self.assertFalse(reps[7])

        reps = self.call_walker("json_file")
        self.assertEqual({"testing": 1}, reps[0])
        self.assertTrue(reps[1])
        self.assertFalse(reps[2])

        reps = self.call_walker("download_file")
        uuid = reps[0]
        self.assertEqual(
            {
                "id": f"{uuid}",
                "name": "jaseci_bible.pdf",
                "content_type": "application/pdf",
                "field": None,
                "absolute_name": f"{uuid}-jaseci_bible.pdf",
                "absolute_path": f"/tmp/{uuid}-jaseci_bible.pdf",
                "persist": False,
            },
            reps[1],
        )

        reps = self.mast._h.get_file_handler(
            self.call_walker("return_any_file", False)["report_file"]
        ).attr()
        uuid = reps["id"]
        self.assertEqual(
            {
                "id": f"{uuid}",
                "name": "jaseci_bible.pdf",
                "content_type": "application/pdf",
                "field": None,
                "absolute_name": f"{uuid}-jaseci_bible.pdf",
                "absolute_path": f"/tmp/{uuid}-jaseci_bible.pdf",
                "persist": False,
            },
            reps,
        )

        reps = self.mast._h.get_file_handler(
            self.call_walker("return_any_file_with_name", False)["report_file"]
        ).attr()
        uuid = reps["id"]
        self.assertEqual(
            {
                "id": f"{uuid}",
                "name": "jaseci_bible.pdf",
                "content_type": "application/pdf",
                "field": None,
                "absolute_name": f"{uuid}-jaseci_bible.pdf",
                "absolute_path": f"/tmp/{uuid}-jaseci_bible.pdf",
                "persist": False,
            },
            reps,
        )

    @patch("requests.post")
    def test_manual_upload_file(self, mock_post):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("file.jac")}],
        )

        self.call_walker("manual_upload_file")
        self.assertTrue(mock_post.called)
        args, kwargs = mock_post.call_args
        self.assertEqual(1, len(args))
        self.assertEqual("https://sample/com", args[0])
        self.assertEqual({"test": 1}, kwargs["data"])
        self.assertEqual({"test": 2}, kwargs["headers"])
        self.assertEqual("file", kwargs["files"][0][0])
        self.assertEqual("jaseci_bible.pdf", kwargs["files"][0][1][0])
        self.assertIsInstance(kwargs["files"][0][1][1], BufferedReader)
        self.assertEqual("application/pdf", kwargs["files"][0][1][2])



from jaseci.utils.test_core import CoreTest, jac_testcase


class DateTest(CoreTest):
    fixture_src = __file__

    @jac_testcase("date.jac", "sample")
    def test_date_syntax(self, ret):
        self.assertEqual(
            ret["report"],
            [
                {
                    "day": 26,
                    "hour": 7,
                    "microsecond": 93606,
                    "minute": 42,
                    "month": 4,
                    "second": 53,
                    "weekday": 3,
                    "year": 2023,
                    "iso": "2023-04-26T07:42:53.093606",
                },
                "2023-06-05T07:42:53.093606",
                "2023-04-26T00:42:53.093606",
            ],
        )



"""
Queue api functions as a mixin
"""

from jaseci.extens.api.interface import Interface
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.task_svc import TaskService


class QueueApi:
    """
    Queue APIs

    APIs used for celery configuration and monitoring
    """

    @Interface.private_api(allowed_methods=["get"])
    def walker_queue_check(self, task_id: str = ""):
        """
        Monitor Queues. Check the task if it's still
            pending, running or already have the result
        """
        task = JsOrc.svc("task", TaskService)
        if not task.is_running():
            return "Task hook is not yet initialized!"

        if not task_id:
            return task.inspect_tasks()
        else:
            return task.get_by_task_id(task_id)

    @Interface.private_api(allowed_methods=["get"])
    def walker_queue_wait(self, task_id: str, timeout: int = 30):
        """
        Forcely wait the queues until executed.

        :param task_id: the task need to wait for result
        :param timeout: force timeout on your current request
            to avoid hang up if there's too many task running currently
        """
        task = JsOrc.svc("task", TaskService)
        if not task.is_running():
            return "Task hook is not yet initialized!"

        if not task_id:
            return "Task id is required!"
        else:
            return task.get_by_task_id(task_id, True, timeout)

    @Interface.private_api()
    def add_scheduled_walker(self, name: str, schedule: dict, body: dict = {}):
        """
        Create a scheduled walker

        :param name: name of the actual queue for reference.
        :param schedule: can be interval or cron.
            Format: {"type": {% interval or cron %}, "conf": {% conf %}, "one_off": {% true if one time only; default false %} }
            Interval conf: {"every": {% integer %}, "period": {% days | hours | minutes | seconds | microseconds %}
            cron conf: {"minute": "*", "hour": "*", "day_of_week": "*", "day_of_month": "*", "month_of_year": "*"}
        :param body: your periodic_task kwargs (dict)
            mst: requestor's master or superadmin requested master - defaults to requestor
            wlk: target walker - defaults to init
            ctx: context to be used upon call - defaults to {}
            nd: target node - defaults to root
            snt: preferred sentinel - defaults to master's active sentinel
        """
        task = JsOrc.svc("task", TaskService)
        if not task.is_running():
            return "Task hook is not yet initialized!"

        from jaseci.prim.super_master import SuperMaster

        if not isinstance(self, SuperMaster) or not body.get("mst"):
            body["mst"] = self.jid

        return (
            "Scheduled Walker created successfully!"
            if task.add_scheduled_queue(task.scheduled_walker, name, schedule, body)
            else "Django is required to support scheduled walker!"
        )

    @Interface.admin_api()
    def add_scheduled_sequence(self, name: str, schedule: dict, body: dict):
        """
        Create a scheduled sequence. More like call multiple api in order.

        :param name: name of the actual queue for reference.
        :param schedule: can be interval or cron.
            Format: {"type": {% interval or cron %}, "conf": {% conf %}, "one_off": {% true if one time only; default false %} }
            Interval conf: {"every": {% integer %}, "period": {% days | hours | minutes | seconds | microseconds %}
            cron conf: {"minute": "*", "hour": "*", "day_of_week": "*", "day_of_month": "*", "month_of_year": "*"}
        :param body: refer to jaseci/docs/docs/deployment/extension_services/task.md
            under schedule sequence argument structure
        """
        task = JsOrc.svc("task", TaskService)
        if not task.is_running():
            return "Task hook is not yet initialized!"

        return (
            "Scheduled Sequence created successfully!"
            if task.add_scheduled_queue(task.scheduled_sequence, name, schedule, body)
            else "Django is required to support scheduled walker!"
        )

    @Interface.private_api()
    def get_scheduled_queues(
        self, limit: int = 10, offset: int = 0, asc: bool = False, search: str = None
    ):
        """
        Get all periodic_task

        :param limit: result limit
        :param offset: query offset
        :param asc: sorting
        :param search: search by name
        """
        task = JsOrc.svc("task", TaskService)
        if not task.is_running():
            return "Task hook is not yet initialized!"

        return task.get_scheduled_queues(limit, offset, asc, search, self)

    @Interface.private_api()
    def delete_scheduled_queue(self, scheduled_queue_id: int):
        """
        Delete specific periodic_task

        :param scheduled_queue_id: the id of periodic task to delete
        """
        task = JsOrc.svc("task", TaskService)
        if not task.is_running():
            return "Task hook is not yet initialized!"

        result = task.delete_scheduled_queue(scheduled_queue_id, self)
        if result == None:
            return "Django is required to support scheduled walker!"
        return (
            "Successfully deleted!"
            if result
            else f"Scheduled Queue with id({scheduled_queue_id}) is not existing!"
        )



"""
Action class for Jaseci

Each action has an id, name, timestamp and it's set of edges.
"""

from jaseci.prim.element import Element
from jaseci.jsorc.live_actions import live_actions
from jaseci.jac.jac_set import JacSet
import inspect
import time

from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.actions.actions_manager import ActionManager
from jaseci.jac.ir.jac_code import JacCode
from jaseci.jac.interpreter.interp import Interp
from jaseci.jac.machine.jac_scope import JacScope


class Ability(Element, JacCode, Interp):
    """
    Abilities class for Jaseci
    """

    def __init__(
        self, code_ir=None, preset_in_out=None, access_list=None, *args, **kwargs
    ):
        self.preset_in_out = preset_in_out  # Not using _ids convention
        self.access_list = access_list
        Element.__init__(self, *args, **kwargs)
        JacCode.__init__(self, code_ir=code_ir)
        Interp.__init__(self)

    def run_ability(self, here, visitor):
        """
        Run ability
        """
        self.push_scope(
            JacScope(
                parent=self,
                name=f"a_run:{self.get_jac_ast().loc_str()}",
                has_obj=here,
                here=here,
                visitor=visitor,
            )
        )
        self.run_code_block(self.get_jac_ast())
        self.pop_scope()

    def run_action(self, param_list, scope, interp, jac_ast):
        """
        param_list should be passed as list of values to lib functions
        Also note that Jac stores preset_in_out as input/output list of hex
        ids since preset_in_out doesn't use _ids convention
        """
        action_name = self.name
        if not interp.check_builtin_action(action_name):
            interp.rt_error(f"Cannot execute {action_name} - Not Found", jac_ast)
            return None
        func = live_actions[action_name]
        args = inspect.getfullargspec(func)
        self.do_auto_conversions(args, param_list)
        args = args[0] + args[4]

        action_manager = JsOrc.get("action_manager", ActionManager)
        action_manager.pre_action_call_hook()

        ts = time.time()
        try:
            if "meta" in args:
                param_list["kwargs"]["meta"] = {
                    "m_id": scope.parent._m_id,
                    "h": scope.parent._h,
                    "scope": scope,
                    "interp": interp,
                }
            result = func(*param_list["args"], **param_list["kwargs"])
        except Exception as e:
            interp.rt_error(e, jac_ast, True)
        t = time.time() - ts
        action_manager.post_action_call_hook(action_name, t)
        return result

    def do_auto_conversions(self, args, params):
        """
        Automatically make conversions for jac to internal, e.g., list to jac_set
        """
        for i in args.annotations.keys():
            if args.annotations[i] == JacSet:
                idx = args.args.index(i)
                if idx < len(params["args"]):
                    params["args"][idx] = JacSet(in_list=params["args"][idx])
                if i in params["kwargs"]:
                    params["kwargs"][i] = JacSet(in_list=params["kwargs"][i])



"""Built in actions for Jaseci"""

from jaseci.jsorc.live_actions import jaseci_action
import os
import json
import base64


@jaseci_action()
def load_str(fn: str, max_chars: int = None):
    """Standard built in for loading from file to string"""
    with open(fn, "r") as file:
        data = file.read(max_chars)
    return data


@jaseci_action()
def load_json(fn: str):
    """Standard built in for loading json from file to dictionary"""
    with open(fn, "r") as file:
        data = json.load(file)
    return data


@jaseci_action()
def load_to_b64(fn: str):
    """Standard built in for loading binary from file to base64"""
    with open(fn, "rb") as file:
        data = base64.b64encode(file.read()).decode("ascii")
    return data


@jaseci_action()
def dump_str(fn: str, s: str):
    """Standard built in for dumping to file from string"""
    with open(fn, "w") as file:
        num_chars = file.write(s)
    return num_chars


@jaseci_action()
def dump_json(fn: str, obj, indent: int = None):
    """Standard built in for dumping json to file from dictionary"""
    with open(fn, "w") as file:
        json.dump(obj, file, indent=indent)


@jaseci_action()
def dump_from_b64(fn: str, b64: str):
    """Standard built in for dumping binary to file from from base64"""
    with open(fn, "wb") as file:
        file.write(base64.b64decode(b64.encode("ascii")))


@jaseci_action()
def append_str(fn: str, s: str):
    """Standard built in for appending to file from string"""
    with open(fn, "a") as file:
        num_chars = file.write(s)
    return num_chars


@jaseci_action()
def append_from_b64(fn: str, b64: str):
    """Standard built in for appending binary to file from from base64"""
    with open(fn, "ab") as file:
        file.write(base64.b64decode(b64.encode("ascii")))


@jaseci_action()
def delete(fn: str):
    """Standard built in for deleting a file"""
    if os.path.exists(fn):
        os.remove(fn)
        return True
    else:
        return False



import math

from jaseci.jsorc.live_actions import jaseci_action


@jaseci_action()
def ceil(x):
    # return int
    return math.ceil(x)


@jaseci_action()
def comb(n, k):
    # return int
    return math.comb(n, k)


@jaseci_action()
def copysign(x, y):
    # return float
    return math.copysign(x, y)


@jaseci_action()
def fabs(x):
    # return float
    return math.fabs(x)


@jaseci_action()
def factorial(n):
    # return int
    return math.factorial(n)


@jaseci_action()
def floor(x):
    # return int
    return math.floor(x)


@jaseci_action()
def fmod(x, y):
    # return float
    return math.fmod(x, y)


@jaseci_action()
def frexp(x):
    # return tuple
    _tuple = math.frexp(x)
    return list(_tuple)


@jaseci_action()
def fsum(iterable: list):
    # return float
    return math.fsum(iterable)


@jaseci_action()
def gcd(args: list):
    args = tuple(args)
    # return int
    return math.gcd(*args)


@jaseci_action()
def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    # return boolean
    return math.isclose(a, b, rel_tol=rel_tol, abs_tol=abs_tol)


@jaseci_action()
def isfinite(x):
    # return boolean
    return math.isfinite(x)


@jaseci_action()
def isinf(x):
    # return boolean
    return math.isinf(x)


@jaseci_action()
def isnan(x):
    # return boolean
    return math.isnan(x)


@jaseci_action()
def isqrt(n):
    # return int
    return math.isqrt(n)


@jaseci_action()
def lcm(args: list):
    args = tuple(args)
    # return int
    return math.lcm(*args)


@jaseci_action()
def ldexp(x, i):
    # return float
    return math.ldexp(x, i)


@jaseci_action()
def modf(x):
    _tupple = math.modf(x)
    # return list
    return list(_tupple)


@jaseci_action()
def nextafter(x, y):
    # return float
    return math.nextafter(x, y)


@jaseci_action()
def perm(n, k=None):
    # return int
    return math.perm(n, k)


@jaseci_action()
def prod(iterable, start=1):
    return math.prod(iterable, start=start)


@jaseci_action()
def remainder(x, y):
    # return float
    return math.remainder(x, y)


@jaseci_action()
def trunc(x):
    # return int
    return math.trunc(x)


@jaseci_action()
def ulp(x):
    # return float
    return math.ulp(x)


@jaseci_action()
def cubert(x):
    # return float
    return x ** (1 / 3)


@jaseci_action()
def exp(x):
    # return float
    return math.exp(x)


@jaseci_action()
def log(x, base):
    # return float
    return math.log(x, base)



"""Built in actions for Jaseci"""

import mimetypes
import magic
from os import path
from urllib.parse import urlparse
from requests import get
from jaseci.jsorc.live_actions import jaseci_action


@jaseci_action(act_group=["fh"])
def load(path: str, meta: dict = {}):
    """temp"""
    from jaseci.utils.file_handler import FileHandler

    return meta["h"].add_file_handler(FileHandler.fromPath(path))


@jaseci_action(act_group=["fh"])
def new(
    name: str,
    content_type: str = None,
    field: str = None,
    persist: bool = False,
    meta: dict = {},
):
    """temp"""
    from jaseci.utils.file_handler import FileHandler

    return meta["h"].add_file_handler(
        FileHandler(name=name, content_type=content_type, field=field, persist=persist)
    )


@jaseci_action
def guess_type(filename: str) -> tuple:
    """temp"""
    return mimetypes.guess_type(filename)


@jaseci_action(act_group=["fh"])
def update(
    id: str,
    name: str = None,
    content_type: str = None,
    field: str = None,
    persist: bool = None,
    meta: dict = {},
):
    """temp"""
    from jaseci.utils.file_handler import FileHandler

    file_handler: FileHandler = meta["h"].get_file_handler(id)
    if name:
        file_handler.name = name
        if not content_type:
            file_handler.content_type = mimetypes.guess_type(name)[0]

    if content_type:
        file_handler.content_type = content_type

    if field:
        file_handler.field = field

    if persist != None:
        file_handler.persist = persist

    return file_handler.attr()


@jaseci_action(act_group=["fh"])
def read(
    id: str, offset: int = None, mode: str = "r", encoding: str = None, meta: dict = {}
):
    """temp"""
    return meta["h"].get_file_handler(id).read(offset, mode, encoding)


@jaseci_action(act_group=["fh"])
def seek(id: str, offset: int, whence: int = 0, meta: dict = {}):
    """temp"""
    return meta["h"].get_file_handler(id).seek(offset, whence)


@jaseci_action(act_group=["fh"])
def open(id: str, mode: str = "r", encoding: str = None, meta: dict = {}, **kwargs):
    """temp"""
    meta["h"].get_file_handler(id).open(mode, encoding, False, **kwargs)


@jaseci_action(act_group=["fh"])
def is_open(id: str, meta: dict = {}):
    """temp"""
    return meta["h"].get_file_handler(id).is_open()


@jaseci_action(act_group=["fh"])
def exists(id: str, meta: dict = {}):
    """temp"""
    return meta["h"].get_file_handler(id).exists()


@jaseci_action(act_group=["fh"])
def write(id: str, content: str, meta: dict = {}):
    """temp"""
    meta["h"].get_file_handler(id).write(content)


@jaseci_action(act_group=["fh"])
def flush(id: str, meta: dict = {}):
    """temp"""
    meta["h"].get_file_handler(id).flush()


@jaseci_action(act_group=["fh"])
def close(id: str, meta: dict = {}):
    """temp"""
    meta["h"].get_file_handler(id).close()


@jaseci_action(act_group=["fh"])
def detach(id: str, persist: bool = None, meta: dict = {}):
    """temp"""
    from jaseci.utils.file_handler import FileHandler

    file_handler: FileHandler = meta["h"].pop_file_handler(id)
    file_info = file_handler.attr()

    if (persist is None and not file_handler.persist) or persist is False:
        file_handler.delete()
    else:
        file_handler.close()

    return file_info


@jaseci_action(act_group=["fh"])
def delete(id: str, meta: dict = {}):
    """temp"""
    meta["h"].pop_file_handler(id).delete()


@jaseci_action(act_group=["fh"])
def attr(id: str, meta: dict = {}):
    """temp"""
    return meta["h"].get_file_handler(id).attr()


@jaseci_action(act_group=["fh"])
def to_bytes(id: str, meta: dict = {}):
    """temp"""
    return meta["h"].get_file_handler(id).to_bytes()


@jaseci_action(act_group=["fh"])
def to_base64(id: str, meta: dict = {}):
    """temp"""
    return meta["h"].get_file_handler(id).to_base64()


@jaseci_action(act_group=["fh"])
def to_json(id: str, meta: dict = {}):
    """temp"""
    return meta["h"].get_file_handler(id).to_json()


@jaseci_action(act_group=["fh"])
def dump_json(id: str, json, indent: int = None, meta: dict = {}):
    """temp"""
    meta["h"].get_file_handler(id).dump_json(json, indent)


@jaseci_action(act_group=["fh"])
def download(url: str, header: dict = {}, meta: dict = {}):
    """Standard built in for download file from url"""
    from jaseci.utils.file_handler import FileHandler

    tmp = FileHandler(path.basename(urlparse(url).path) or "file")
    meta["h"].add_file_handler(tmp)

    with get(url, stream=True, headers=header) as res:
        res.raise_for_status()
        tmp.open("wb+")
        for chunk in res.iter_content(chunk_size=8192):
            tmp.buffer.write(chunk)

        tmp.content_type = magic.from_buffer(tmp.read(0), mime=True)
        tmp.close()

    return tmp.id
